#include <stdbool.h>
#include <stddef.h>
#include <stdint.h>

#include "kprintf.h"
#include "memory.h"
#include "utils.h"

#define MAX_FRAMES 1048319

static uint32_t end_kernel; // WIP but it's useless now

static uint8_t frame_table[MAX_FRAMES];
static uint32_t remaining_frames = MAX_FRAMES;

void *alloc_frames(size_t size)
{
	const uint32_t nb_frames = CEIL(size, PAGE_SIZE);
	if (nb_frames > remaining_frames) {
		kprintf(KERN_CRIT "Not enough frames (max: %d)\n", MAX_FRAMES);
		return NULL;
	}

	size_t i = 0;
	while (i < MAX_FRAMES) {
		size_t free_frames = 1;
		while (!frame_table[i + free_frames] && free_frames < nb_frames)
			free_frames++;
		if (free_frames == nb_frames)
			goto end;
		if (!free_frames)
			i++;
		i += free_frames;
	}
end:
	if (i != MAX_FRAMES) {
		for (size_t j = 0; j < nb_frames; j++)
			frame_table[j + i] = 1;
		remaining_frames -= nb_frames;
		return &end_kernel + i * PAGE_SIZE;
	}
	kprintf(KERN_WARNING "Not enough frames available\n", MAX_FRAMES);
	return NULL;
}

int free_frames(void *frame_ptr, size_t size)
{
	const uint32_t nb_frames = CEIL(size, PAGE_SIZE);
	const uint32_t start = (frame_ptr - (void *)&end_kernel) / PAGE_SIZE;

	if (start > MAX_FRAMES || frame_ptr < (void *)&end_kernel) {
		kprintf(KERN_WARNING "Address out of range\n");
		return -1;
	} else if ((uint32_t)frame_ptr % PAGE_SIZE) {
		kprintf(KERN_WARNING "Invalid address\n");
		return -1;
	} else if (start + nb_frames > MAX_FRAMES) {
		kprintf(KERN_WARNING "Invalid number of frames\n");
		return -1;
	}
	for (size_t i = start; i < start + nb_frames; i++)
		frame_table[i] = 0;
	remaining_frames += nb_frames;
	return 0;
}
